

set -ex



pip check
pytest -vv -ra --showlocals -m "not (isolated or network or virtualenv or setuptools or broken_on_urct)" --deselect=tests/test_editable_unit.py::test_navigate_editable_pkg -k "not (sdist_time_hash_set_epoch or sdist_hash)"
exit 0
